# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Standard implementation of message command argument parsing."""
from __future__ import annotations

__all__: list[str] = [
    "AbstractOptionParser",
    "Argument",
    "ConverterSig",
    "Option",
    "Parameter",
    "ShlexParser",
    "with_argument",
    "with_greedy_argument",
    "with_multi_argument",
    "with_multi_option",
    "with_option",
    "with_parser",
]

import abc
import asyncio
import copy
import itertools
import shlex
import typing
from collections import abc as collections

import typing_extensions

from . import abc as tanjun
from . import conversion
from . import errors

if typing.TYPE_CHECKING:
    from typing_extensions import Self

    _CommandT = typing.TypeVar("_CommandT", bound=tanjun.MessageCommand[typing.Any])
    _T_contra = typing.TypeVar("_T_contra", contravariant=True)

    class _CmpProto(typing.Protocol[_T_contra]):
        def __gt__(self, __other: _T_contra) -> bool:
            raise NotImplementedError

        def __lt__(self, __other: _T_contra) -> bool:
            raise NotImplementedError

    class _SizedCmpProto(_CmpProto[_T_contra]):
        def __len__(self) -> int:
            raise NotImplementedError

    _CmpProtoT = typing.TypeVar("_CmpProtoT", bound=_CmpProto[typing.Any])
    _MaybeIterable = typing.Union[collections.Iterable["_T"], "_T"]
    _SizedCmpProtoT = typing.TypeVar("_SizedCmpProtoT", bound=_SizedCmpProto[typing.Any])


_T = typing.TypeVar("_T")

# 3.9 and 3.10 just can't handle ending Concatenate with ... so we lie about this at runtime.
if typing.TYPE_CHECKING:
    _P = typing_extensions.ParamSpec("_P")
    _ConverterSig = collections.Callable[
        typing_extensions.Concatenate[str, _P], typing.Union[collections.Coroutine[typing.Any, typing.Any, _T], _T]
    ]

    ConverterSig = _ConverterSig[..., _T]
    """Type hint of a converter used within a parser instance.

    This represents the signatures `def (str, ...) -> Any` and
    `async def (str, ...) -> Any` where dependency injection is supported.
    """

else:
    import types

    ConverterSig = types.GenericAlias(
        collections.Callable[..., typing.Union[collections.Coroutine[typing.Any, typing.Any, _T], _T]], (_T,)
    )


UndefinedT = typing.Literal[tanjun.NO_DEFAULT]
"""Deprecated alias of `typing.Literal[tanjun.abc.NO_DEFAULT]`."""

UndefinedDefaultT = typing.Literal[tanjun.NO_DEFAULT]
"""Deprecated alias of `typing.Literal[tanjun.abc.NO_DEFAULT]`."""

UNDEFINED = tanjun.NO_DEFAULT
"""Deprecated alias of [tanjun.abc.NO_DEFAULT][]."""

UNDEFINED_DEFAULT = tanjun.NO_DEFAULT
"""Deprecated alias of [tanjun.abc.NO_DEFAULT][]."""


class AbstractOptionParser(tanjun.MessageParser, abc.ABC):
    """Abstract interface of a message content parser."""

    __slots__ = ()

    @property
    @abc.abstractmethod
    def arguments(self) -> collections.Sequence[Argument]:
        """Sequence of the positional arguments registered with this parser."""

    @property
    @abc.abstractmethod
    def options(self) -> collections.Sequence[Option]:
        """Sequence of the named options registered with this parser."""

    @typing.overload
    @abc.abstractmethod
    def add_argument(
        self,
        key: str,
        /,
        converters: _MaybeIterable[ConverterSig[typing.Any]],
        *,
        default: typing.Any = tanjun.NO_DEFAULT,
        greedy: bool = False,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    @abc.abstractmethod
    def add_argument(
        self,
        key: str,
        /,
        converters: _MaybeIterable[ConverterSig[str]] = (),
        *,
        default: typing.Any = tanjun.NO_DEFAULT,
        greedy: bool = False,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_CmpProto[str]] = None,
        max_value: typing.Optional[_CmpProto[str]] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    @abc.abstractmethod
    def add_argument(
        self,
        key: str,
        /,
        converters: _MaybeIterable[ConverterSig[_SizedCmpProtoT]],
        *,
        default: typing.Any = tanjun.NO_DEFAULT,
        greedy: bool = False,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_SizedCmpProtoT] = None,
        max_value: typing.Optional[_SizedCmpProtoT] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    @abc.abstractmethod
    def add_argument(
        self,
        key: str,
        /,
        converters: _MaybeIterable[ConverterSig[collections.Sized]],
        *,
        default: typing.Any = tanjun.NO_DEFAULT,
        greedy: bool = False,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    @abc.abstractmethod
    def add_argument(
        self,
        key: str,
        /,
        converters: _MaybeIterable[ConverterSig[_CmpProtoT]],
        *,
        default: typing.Any = tanjun.NO_DEFAULT,
        greedy: bool = False,
        min_value: typing.Optional[_CmpProtoT] = None,
        max_value: typing.Optional[_CmpProtoT] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @abc.abstractmethod
    def add_argument(
        self,
        key: str,
        /,
        converters: _MaybeIterable[ConverterSig[typing.Any]] = (),
        *,
        default: typing.Any = tanjun.NO_DEFAULT,
        greedy: bool = False,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_CmpProto[typing.Any]] = None,
        max_value: typing.Optional[_CmpProto[typing.Any]] = None,
        multi: bool = False,
    ) -> Self:
        """Add a positional argument type to the parser..

        !!! note
            Order matters for positional arguments.

        Parameters
        ----------
        key
            The string identifier of this argument (may be used to pass the result
            of this argument to the command's callback during execution).
        converters
            The converter(s) this argument should use to handle values passed to it
            during parsing.

            If no converters are provided then the raw string value will be passed.

            Only the first converter to pass will be used.
        default
            The default value of this argument, if left as
            [tanjun.abc.NO_DEFAULT][] then this will have no default.

            If this is [tanjun.abc.NO_PASS][] then the `key` parameter won't be
            passed when no value was provided.
        greedy
            Whether or not this argument should be greedy (meaning that it
            takes in the remaining argument values).
        min_length
            Assert that a string argument's length is greater than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        max_length
            Assert that a string argument's length is less than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        min_value
            Assert that the parsed value(s) for this argument are greater than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        max_value
            Assert that the parsed value(s) for this argument are less than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        multi
            Whether this argument can be passed multiple times.

        Returns
        -------
        Self
            This parser to enable chained calls.

        Raises
        ------
        ValueError
            If `key` isn't valid for any of the commands this parser is linked
            to where `validate_arg_keys` is [True][].
        """

    @typing.overload
    @abc.abstractmethod
    def add_option(
        self,
        key: str,
        name: str,
        /,
        *names: str,
        converters: _MaybeIterable[ConverterSig[typing.Any]],
        default: typing.Any,
        empty_value: typing.Any = tanjun.NO_DEFAULT,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    @abc.abstractmethod
    def add_option(
        self,
        key: str,
        name: str,
        /,
        *names: str,
        converters: _MaybeIterable[ConverterSig[str]] = (),
        default: typing.Any,
        empty_value: typing.Any = tanjun.NO_DEFAULT,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_CmpProto[str]] = None,
        max_value: typing.Optional[_CmpProto[str]] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    @abc.abstractmethod
    def add_option(
        self,
        key: str,
        name: str,
        /,
        *names: str,
        converters: _MaybeIterable[ConverterSig[_SizedCmpProtoT]],
        default: typing.Any,
        empty_value: typing.Any = tanjun.NO_DEFAULT,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_SizedCmpProtoT] = None,
        max_value: typing.Optional[_SizedCmpProtoT] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    @abc.abstractmethod
    def add_option(
        self,
        key: str,
        name: str,
        /,
        *names: str,
        converters: _MaybeIterable[ConverterSig[collections.Sized]],
        default: typing.Any,
        empty_value: typing.Any = tanjun.NO_DEFAULT,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    @abc.abstractmethod
    def add_option(
        self,
        key: str,
        name: str,
        /,
        *names: str,
        converters: _MaybeIterable[ConverterSig[_CmpProtoT]],
        default: typing.Any,
        empty_value: typing.Any = tanjun.NO_DEFAULT,
        min_value: typing.Optional[_CmpProtoT] = None,
        max_value: typing.Optional[_CmpProtoT] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @abc.abstractmethod
    def add_option(
        self,
        key: str,
        name: str,
        /,
        *names: str,
        converters: _MaybeIterable[ConverterSig[typing.Any]] = (),
        default: typing.Any,
        empty_value: typing.Any = tanjun.NO_DEFAULT,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_CmpProto[typing.Any]] = None,
        max_value: typing.Optional[_CmpProto[typing.Any]] = None,
        multi: bool = False,
    ) -> Self:
        """Add an named option to this parser.

        Parameters
        ----------
        key
            The string identifier of this option which will be used to pass the
            result of this option to the command's callback during execution as
            a keyword argument.
        name
            The name of this option used for identifying it in the parsed content.
        *names
            Other names of this option used for identifying it in the parsed content.
        default
            The default value of this option, unlike arguments this is required
            for options.
        converters
            The converter(s) this option should use to handle values passed to it
            during parsing.

            If no converters are provided then the raw string value will be passed.

            Only the first converter to pass will be used.
        empty_value
            The value to use if this option is provided without a value.

            If left as [tanjun.abc.NO_DEFAULT][] then this option
            will error if it's provided without a value.

            [tanjun.abc.NO_PASS][] is not supported for this.
        min_length
            Assert that a string argument's length is greater than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        max_length
            Assert that a string argument's length is less than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        min_value
            Assert that the parsed value(s) for this option are greater than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        max_value
            Assert that the parsed value(s) for this option are less than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        multi
            If this option can be provided multiple times.

        Returns
        -------
        Self
            This parser to enable chained calls.

        Raises
        ------
        ValueError
            If `key` isn't valid for any of the commands this parser is linked
            to where `validate_arg_keys` is [True][].
        """


@typing_extensions.deprecated("Use AbstractOptionParser")
class AbstractParser(AbstractOptionParser):
    """Deprecated alias of [AbstractOptionParser][tanjun.parsing.AbstractOptionParser]."""

    __slots__ = ()


class _ShlexTokenizer:
    __slots__ = ("__arg_buffer", "__last_name", "__options_buffer", "__shlex")

    def __init__(self, content: str, /) -> None:
        self.__arg_buffer: list[str] = []
        self.__last_name: typing.Optional[str] = None
        self.__options_buffer: list[tuple[str, typing.Optional[str]]] = []
        self.__shlex = shlex.shlex(content, posix=True)
        self.__shlex.commenters = ""
        self.__shlex.quotes = '"'
        self.__shlex.whitespace = " "
        self.__shlex.whitespace_split = True

    def collect_raw_options(self) -> collections.Mapping[str, collections.Sequence[typing.Optional[str]]]:
        results: dict[str, list[typing.Optional[str]]] = {}

        while (option := self.next_raw_option()) is not None:
            name, value = option

            if name not in results:
                results[name] = []

            results[name].append(value)

        return results

    def iter_raw_arguments(self) -> collections.Iterator[str]:
        while (argument := self.next_raw_argument()) is not None:
            yield argument

    def next_raw_argument(self) -> typing.Optional[str]:
        if self.__arg_buffer:
            return self.__arg_buffer.pop(0)

        while (value := self.__seek_shlex()) and value[0] == 1:
            self.__options_buffer.append(value[1])

        return value[1] if value else None

    def next_raw_option(self) -> typing.Optional[tuple[str, typing.Optional[str]]]:
        if self.__options_buffer:
            return self.__options_buffer.pop(0)

        while (value := self.__seek_shlex()) and value[0] == 0:
            self.__arg_buffer.append(value[1])

        return value[1] if value else None

    def __seek_shlex(
        self,
    ) -> typing.Union[tuple[typing.Literal[0], str], tuple[typing.Literal[1], tuple[str, typing.Optional[str]]], None]:
        option_name = self.__last_name

        try:
            value = next(self.__shlex)

        except StopIteration:
            if option_name is not None:
                self.__last_name = None
                return (1, (option_name, None))

            return None

        except ValueError as exc:
            raise errors.ParserError(str(exc), None) from exc

        is_option = value.startswith("-")
        if is_option and option_name is not None:
            self.__last_name = value
            return (1, (option_name, None))

        if is_option:
            self.__last_name = value
            return self.__seek_shlex()

        if option_name:
            self.__last_name = None
            return (1, (option_name, value))

        return (0, value)


async def _covert_option_or_empty(
    ctx: tanjun.MessageContext, option: Option, value: typing.Optional[typing.Any], /
) -> typing.Any:
    if value is not None:
        return await option.convert(ctx, value)

    if option.empty_value is not tanjun.NO_DEFAULT:
        return option.empty_value

    raise errors.NotEnoughArgumentsError(f"Option '{option.key} cannot be empty.", option.key)


class _SemanticShlex(_ShlexTokenizer):
    __slots__ = ("__arguments", "__ctx", "__options")

    def __init__(
        self,
        ctx: tanjun.MessageContext,
        arguments: collections.Sequence[Argument],
        options: collections.Sequence[Option],
        /,
    ) -> None:
        super().__init__(ctx.content)
        self.__arguments = arguments
        self.__ctx = ctx
        self.__options = options

    async def parse(self) -> dict[str, typing.Any]:
        raw_options = self.collect_raw_options()
        kwargs: dict[str, typing.Any] = {}

        for option in self.__options:
            values_iter = itertools.chain.from_iterable(
                raw_options[name] for name in option.names if name in raw_options
            )
            if option.is_multi and (values := list(values_iter)):
                kwargs[option.key] = await asyncio.gather(
                    *(_covert_option_or_empty(self.__ctx, option, value) for value in values)
                )

            elif not option.is_multi and (value := next(values_iter, ...)) is not ...:
                if next(values_iter, ...) is not ...:
                    raise errors.TooManyArgumentsError(
                        f"Option `{option.key}` can only take a single value", option.key
                    )

                kwargs[option.key] = await _covert_option_or_empty(self.__ctx, option, value)

            elif option.default is tanjun.NO_DEFAULT:
                # If this is reached then no value was found.
                raise errors.NotEnoughArgumentsError(f"Missing required option `{option.key}`", option.key)

            elif option.default is not tanjun.NO_PASS:
                kwargs[option.key] = option.default

        for argument in self.__arguments:
            if argument.is_greedy and (value := " ".join(self.iter_raw_arguments())):
                kwargs[argument.key] = await argument.convert(self.__ctx, value)

            elif argument.is_multi and (values := list(self.iter_raw_arguments())):
                kwargs[argument.key] = await asyncio.gather(*(argument.convert(self.__ctx, value) for value in values))

            # If the previous two statements failed on getting raw arguments then this will as well.
            elif (optional_value := self.next_raw_argument()) is not None:
                kwargs[argument.key] = await argument.convert(self.__ctx, optional_value)

            elif argument.default is tanjun.NO_DEFAULT:
                # If this is reached then no value was found.
                raise errors.NotEnoughArgumentsError(
                    f"Missing value for required argument '{argument.key}'", argument.key
                )

            elif argument.default is not tanjun.NO_PASS:
                kwargs[argument.key] = argument.default

            if argument.is_greedy or argument.is_multi:
                break  # Multi and Greedy parameters should always be the last parameter.

        return kwargs


def _get_or_set_parser(command: tanjun.MessageCommand[typing.Any], /) -> AbstractOptionParser:
    if not command.parser:
        parser = ShlexParser()
        command.set_parser(parser)
        return parser

    if isinstance(command.parser, AbstractOptionParser):
        return command.parser

    raise TypeError("Expected parser to be an instance of tanjun.parsing.AbstractOptionParser")


@typing.overload
def with_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[typing.Any]],
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    greedy: bool = False,
    multi: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[str]] = (),
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    greedy: bool = False,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_CmpProto[str]] = None,
    max_value: typing.Optional[_CmpProto[str]] = None,
    multi: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[_SizedCmpProtoT]],
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    greedy: bool = False,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_SizedCmpProtoT] = None,
    max_value: typing.Optional[_SizedCmpProtoT] = None,
    multi: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[collections.Sized]],
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    greedy: bool = False,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    multi: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[_CmpProtoT]],
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    greedy: bool = False,
    min_value: typing.Optional[_CmpProtoT] = None,
    max_value: typing.Optional[_CmpProtoT] = None,
    multi: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


def with_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[typing.Any]] = (),
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    greedy: bool = False,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_CmpProto[typing.Any]] = None,
    max_value: typing.Optional[_CmpProto[typing.Any]] = None,
    multi: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    """Add an argument to a message command through a decorator call.

    !!! warning
        Since order matters for positional arguments, you'll want to keep in
        mind that decorator execution starts at the decorator closest to the
        command and goes upwards with this deciding where a positional
        argument is located in a command's signature.

    !!! note
        If no parser is explicitly set on the command this is decorating before
        this decorator call then this will set
        [ShlexParser][tanjun.parsing.ShlexParser] as the parser.

    Examples
    --------
    ```python
    import tanjun

    @tanjun.parsing.with_argument("command", converters=int, default=42)
    @tanjun.parsing.with_parser
    @tanjun.component.as_message_command("command")
    async def command(self, ctx: tanjun.abc.Context, /, argument: int):
        ...
    ```

    Parameters
    ----------
    key
        The string identifier of this argument (may be used to pass the result
        of this argument to the command's callback during execution).
    converters
        The converter(s) this argument should use to handle values passed to it
        during parsing.

        If no converters are provided then the raw string value will be passed.

        Only the first converter to pass will be used.
    default
        The default value of this argument, if left as
        [tanjun.abc.NO_DEFAULT][] then this will have no default.

        If this is [tanjun.abc.NO_PASS][] then the `key` parameter won't be
        passed when no value was provided.
    greedy
        Whether or not this argument should be greedy (meaning that it
        takes in the remaining argument values).
    min_length
        Assert that a string argument's length is greater than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    max_length
        Assert that a string argument's length is less than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    min_value
        Assert that the parsed value(s) for this argument are greater than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    max_value
        Assert that the parsed value(s) for this argument are less than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.

    multi
        Whether this argument can be passed multiple times.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.MessageCommand], tanjun.abc.MessageCommand]:
        Decorator function for the message command this argument is being added to.

    Raises
    ------
    ValueError
        If `key` isn't valid for any of the commands this command's parser is
        linked to where `validate_arg_keys` is [True][].
    """

    def decorator(command: _CommandT, /) -> _CommandT:
        _get_or_set_parser(command).add_argument(
            key,
            converters=converters,
            default=default,
            greedy=greedy,
            min_length=min_length,
            max_length=max_length,
            min_value=min_value,
            max_value=max_value,
            multi=multi,
        )
        return command

    return decorator


@typing.overload
def with_greedy_argument(
    key: str, /, converters: _MaybeIterable[ConverterSig[typing.Any]], *, default: typing.Any = tanjun.NO_DEFAULT
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_greedy_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[str]] = (),
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_CmpProto[str]] = None,
    max_value: typing.Optional[_CmpProto[str]] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_greedy_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[_SizedCmpProtoT]],
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_SizedCmpProtoT] = None,
    max_value: typing.Optional[_SizedCmpProtoT] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_greedy_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[collections.Sized]],
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_greedy_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[_CmpProtoT]],
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    min_value: typing.Optional[_CmpProtoT] = None,
    max_value: typing.Optional[_CmpProtoT] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


def with_greedy_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[typing.Any]] = (),
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_CmpProto[typing.Any]] = None,
    max_value: typing.Optional[_CmpProto[typing.Any]] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    """Add a greedy argument to a message command through a decorator call.

    A greedy argument will consume the remaining positional arguments and pass
    them through to the converters as one joined string while also requiring
    that at least one more positional argument is remaining unless a default is
    set.

    !!! warning
        Since order matters for positional arguments, you'll want to keep in
        mind that decorator execution starts at the decorator closest to the
        command and goes upwards with this deciding where a positional
        argument is located in a command's signature.

    !!! note
        If no parser is explicitly set on the command this is decorating before
        this decorator call then this will set
        [ShlexParser][tanjun.parsing.ShlexParser] as the parser.

    Examples
    --------
    ```python
    import tanjun

    @tanjun.parsing.with_greedy_argument("command", converters=StringView)
    @tanjun.parsing.with_parser
    @tanjun.component.as_message_command("command")
    async def command(self, ctx: tanjun.abc.Context, /, argument: StringView):
        ...
    ```

    Parameters
    ----------
    key
        The string identifier of this argument (may be used to pass the result
        of this argument to the command's callback during execution).
    converters
        The converter(s) this argument should use to handle values passed to it
        during parsing.

        If no converters are provided then the raw string value will be passed.

        Only the first converter to pass will be used.
    default
        The default value of this argument, if left as
        [tanjun.abc.NO_DEFAULT][] then this will have no default.

        If this is [tanjun.abc.NO_PASS][] then the `key` parameter won't be
        passed when no value was provided.

        If any converters are provided then this should be compatible
        with the result of them.
    min_length
        Assert that a string argument's length is greater than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    max_length
        Assert that a string argument's length is less than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    min_value
        Assert that the parsed value(s) for this argument are greater than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    max_value
        Assert that the parsed value(s) for this argument are less than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.MessageCommand], tanjun.abc.MessageCommand]:
        Decorator function for the message command this argument is being added to.

    Raises
    ------
    ValueError
        If `key` isn't valid for any of the commands this command's parser is
        linked to where `validate_arg_keys` is [True][].
    """
    return with_argument(
        key,
        converters=converters,
        default=default,
        greedy=True,
        min_length=min_length,
        max_length=max_length,
        min_value=min_value,
        max_value=max_value,
    )


@typing.overload
def with_multi_argument(
    key: str, /, converters: _MaybeIterable[ConverterSig[typing.Any]], *, default: typing.Any = tanjun.NO_DEFAULT
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_multi_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[str]] = (),
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_CmpProto[str]] = None,
    max_value: typing.Optional[_CmpProto[str]] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_multi_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[_SizedCmpProtoT]],
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_SizedCmpProtoT] = None,
    max_value: typing.Optional[_SizedCmpProtoT] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_multi_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[collections.Sized]],
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_multi_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[_CmpProtoT]],
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    min_value: typing.Optional[_CmpProtoT] = None,
    max_value: typing.Optional[_CmpProtoT] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


def with_multi_argument(
    key: str,
    /,
    converters: _MaybeIterable[ConverterSig[typing.Any]] = (),
    *,
    default: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_CmpProto[typing.Any]] = None,
    max_value: typing.Optional[_CmpProto[typing.Any]] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    """Add a multi-argument to a message command through a decorator call.

    A multi argument will consume the remaining positional arguments and pass
    them to the converters through multiple calls while also requiring that
    at least one more positional argument is remaining unless a default is
    set and passing through the results to the command's callback as a
    sequence.

    !!! warning
        Since order matters for positional arguments, you'll want to keep in
        mind that decorator execution starts at the decorator closest to the
        command and goes upwards with this deciding where a positional
        argument is located in a command's signature.

    !!! note
        If no parser is explicitly set on the command this is decorating before
        this decorator call then this will set
        [ShlexParser][tanjun.parsing.ShlexParser] as the parser.

    Examples
    --------
    ```python
    import tanjun

    @tanjun.parsing.with_multi_argument("command", converters=int)
    @tanjun.parsing.with_parser
    @tanjun.component.as_message_command("command")
    async def command(self, ctx: tanjun.abc.Context, /, argument: collections.abc.Sequence[int]):
        ...
    ```

    Parameters
    ----------
    key
        The string identifier of this argument (may be used to pass the result
        of this argument to the command's callback during execution).
    converters
        The converter(s) this argument should use to handle values passed to it
        during parsing.

        If no converters are provided then the raw string value will be passed.

        Only the first converter to pass will be used.
    default
        The default value of this argument, if left as
        [tanjun.abc.NO_DEFAULT][] then this will have no default.

        If this is [tanjun.abc.NO_PASS][] then the `key` parameter won't be
        passed when no value was provided.
    min_length
        Assert that a string argument's length is greater than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    max_length
        Assert that a string argument's length is less than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    min_value
        Assert that the parsed value(s) for this argument are greater than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    max_value
        Assert that the parsed value(s) for this argument are less than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.MessageCommand], tanjun.abc.MessageCommand]:
        Decorator function for the message command this argument is being added to.

    Raises
    ------
    ValueError
        If `key` isn't valid for any of the commands this command's parser is
        linked to where `validate_arg_keys` is [True][].
    """
    return with_argument(
        key,
        converters=converters,
        default=default,
        min_length=min_length,
        max_length=max_length,
        min_value=min_value,
        max_value=max_value,
        multi=True,
    )


@typing.overload
def with_option(
    key: str,
    name: str,
    /,
    *names: str,
    converters: _MaybeIterable[ConverterSig[typing.Any]],
    default: typing.Any,
    empty_value: typing.Any = tanjun.NO_DEFAULT,
    multi: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_option(
    key: str,
    name: str,
    /,
    *names: str,
    converters: _MaybeIterable[ConverterSig[str]] = (),
    default: typing.Any,
    empty_value: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_CmpProto[str]] = None,
    max_value: typing.Optional[_CmpProto[str]] = None,
    multi: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_option(
    key: str,
    name: str,
    /,
    *names: str,
    converters: _MaybeIterable[ConverterSig[_SizedCmpProtoT]],
    default: typing.Any,
    empty_value: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_SizedCmpProtoT] = None,
    max_value: typing.Optional[_SizedCmpProtoT] = None,
    multi: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_option(
    key: str,
    name: str,
    /,
    *names: str,
    converters: _MaybeIterable[ConverterSig[collections.Sized]],
    default: typing.Any,
    empty_value: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    multi: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_option(
    key: str,
    name: str,
    /,
    *names: str,
    converters: _MaybeIterable[ConverterSig[_CmpProtoT]],
    default: typing.Any,
    empty_value: typing.Any = tanjun.NO_DEFAULT,
    min_value: typing.Optional[_CmpProtoT] = None,
    max_value: typing.Optional[_CmpProtoT] = None,
    multi: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


# TODO: add default getter
def with_option(
    key: str,
    name: str,
    /,
    *names: str,
    converters: _MaybeIterable[ConverterSig[typing.Any]] = (),
    default: typing.Any,
    empty_value: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_CmpProto[typing.Any]] = None,
    max_value: typing.Optional[_CmpProto[typing.Any]] = None,
    multi: bool = False,
) -> collections.Callable[[_CommandT], _CommandT]:
    """Add an option to a message command through a decorator call.

    !!! note
        If no parser is explicitly set on the command this is decorating before
        this decorator call then this will set
        [ShlexParser][tanjun.parsing.ShlexParser] as the parser.

    Examples
    --------
    ```python
    import tanjun

    @tanjun.parsing.with_option("command", converters=int, default=42)
    @tanjun.parsing.with_parser
    @tanjun.component.as_message_command("command")
    async def command(self, ctx: tanjun.abc.Context, /, argument: int):
        ...
    ```

    Parameters
    ----------
    key
        The string identifier of this option which will be used to pass the
        result of this argument to the command's callback during execution as
        a keyword argument.
    name
        The name of this option used for identifying it in the parsed content.
    *names
        Other names of this option used for identifying it in the parsed content.
    default
        The default value of this argument, unlike arguments this is required
        for options.
    converters
        The converter(s) this argument should use to handle values passed to it
        during parsing.

        If no converters are provided then the raw string value will be passed.

        Only the first converter to pass will be used.
    empty_value
        The value to use if this option is provided without a value.

        If left as [tanjun.abc.NO_DEFAULT][] then this option will error if
        it's provided without a value.

        [tanjun.abc.NO_PASS][] is not supported for this.
    min_length
        Assert that a string argument's length is greater than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    max_length
        Assert that a string argument's length is less than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    min_value
        Assert that the parsed value(s) for this option are greater than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    max_value
        Assert that the parsed value(s) for this option are less than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    multi
        If this option can be provided multiple times.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.MessageCommand], tanjun.abc.MessageCommand]:
        Decorator function for the message command this option is being added to.

    Raises
    ------
    ValueError
        If `key` isn't valid for any of the commands this command's parser is
        linked to where `validate_arg_keys` is [True][].
    """

    def decorator(command: _CommandT, /) -> _CommandT:
        _get_or_set_parser(command).add_option(
            key,
            name,
            *names,
            converters=converters,
            default=default,
            empty_value=empty_value,
            min_length=min_length,
            max_length=max_length,
            min_value=min_value,
            max_value=max_value,
            multi=multi,
        )
        return command

    return decorator


@typing.overload
def with_multi_option(
    key: str,
    name: str,
    /,
    *names: str,
    converters: _MaybeIterable[ConverterSig[typing.Any]],
    default: typing.Any,
    empty_value: typing.Any = tanjun.NO_DEFAULT,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_multi_option(
    key: str,
    name: str,
    /,
    *names: str,
    converters: _MaybeIterable[ConverterSig[str]] = (),
    default: typing.Any,
    empty_value: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_CmpProto[str]] = None,
    max_value: typing.Optional[_CmpProto[str]] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_multi_option(
    key: str,
    name: str,
    /,
    *names: str,
    converters: _MaybeIterable[ConverterSig[_SizedCmpProtoT]],
    default: typing.Any,
    empty_value: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_SizedCmpProtoT] = None,
    max_value: typing.Optional[_SizedCmpProtoT] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_multi_option(
    key: str,
    name: str,
    /,
    *names: str,
    converters: _MaybeIterable[ConverterSig[collections.Sized]],
    default: typing.Any,
    empty_value: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


@typing.overload
def with_multi_option(
    key: str,
    name: str,
    /,
    *names: str,
    converters: _MaybeIterable[ConverterSig[_CmpProtoT]],
    default: typing.Any,
    empty_value: typing.Any = tanjun.NO_DEFAULT,
    min_value: typing.Optional[_CmpProtoT] = None,
    max_value: typing.Optional[_CmpProtoT] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    ...


def with_multi_option(
    key: str,
    name: str,
    /,
    *names: str,
    converters: _MaybeIterable[ConverterSig[typing.Any]] = (),
    default: typing.Any,
    empty_value: typing.Any = tanjun.NO_DEFAULT,
    min_length: typing.Optional[int] = None,
    max_length: typing.Optional[int] = None,
    min_value: typing.Optional[_CmpProto[typing.Any]] = None,
    max_value: typing.Optional[_CmpProto[typing.Any]] = None,
) -> collections.Callable[[_CommandT], _CommandT]:
    """Add an multi-option to a command's parser through a decorator call.

    A multi option will consume all the values provided for an option and
    pass them through to the converters as an array of strings while also
    requiring that at least one value is provided for the option unless
    a default is set.

    !!! note
        If no parser is explicitly set on the command this is decorating before
        this decorator call then this will set
        [ShlexParser][tanjun.parsing.ShlexParser] as the parser.

    Examples
    --------
    ```python
    import tanjun

    @tanjun.parsing.with_multi_option("command", converters=int, default=())
    @tanjun.parsing.with_parser
    @tanjun.component.as_message_command("command")
    async def command(self, ctx: tanjun.abc.Context, /, argument: collections.abc.Sequence[int]):
        ...
    ```

    Parameters
    ----------
    key
        The string identifier of this option which will be used to pass the
        result of this argument to the command's callback during execution as
        a keyword argument.
    name
        The name of this option used for identifying it in the parsed content.
    *names
        Other names of this option used for identifying it in the parsed content.
    default
        The default value of this argument, unlike arguments this is required
        for options.
    converters
        The converter(s) this argument should use to handle values passed to it
        during parsing.

        If no converters are provided then the raw string value will be passed.

        Only the first converter to pass will be used.
    empty_value
        The value to use if this option is provided without a value.

        If left as [tanjun.abc.NO_DEFAULT][] then this option will error if
        it's provided without a value.

        [tanjun.abc.NO_PASS][] is not supported for this.
    min_length
        Assert that a string argument's length is greater than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    max_length
        Assert that a string argument's length is less than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    min_value
        Assert that the parsed value(s) for this option are greater than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.
    max_value
        Assert that the parsed value(s) for this option are less than or equal to this.

        If any converters are provided then this should be compatible
        with the result of them.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.MessageCommand], tanjun.abc.MessageCommand]:
        Decorator function for the message command this option is being added to.

    Raises
    ------
    ValueError
        If `key` isn't valid for any of the commands this command's parser is
        linked to where `validate_arg_keys` is [True][].
    """
    return with_option(
        key,
        name,
        *names,
        converters=converters,
        default=default,
        empty_value=empty_value,
        min_length=min_length,
        max_length=max_length,
        min_value=min_value,
        max_value=max_value,
        multi=True,
    )


class Parameter:
    """Base class for parameters for the standard parser(s)."""

    __slots__ = (
        "_client",
        "_component",
        "_converters",
        "_default",
        "_is_multi",
        "_key",
        "_min_length",
        "_max_length",
        "_min_value",
        "_max_value",
    )

    def __init__(
        self,
        key: str,
        /,
        *,
        converters: _MaybeIterable[ConverterSig[typing.Any]] = (),
        default: typing.Any = tanjun.NO_DEFAULT,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_CmpProto[typing.Any]] = None,
        max_value: typing.Optional[_CmpProto[typing.Any]] = None,
        multi: bool = False,
    ) -> None:
        """Initialise a parameter."""
        self._client: typing.Optional[tanjun.Client] = None
        self._component: typing.Optional[tanjun.Component] = None
        self._converters: list[ConverterSig[typing.Any]] = []
        self._default = default
        self._is_multi = multi
        self._key = key
        self._min_length = min_length
        self._max_length = max_length
        self._min_value = min_value
        self._max_value = max_value

        if key.startswith("-"):
            raise ValueError("parameter key cannot start with `-`")

        if isinstance(converters, collections.Iterable):
            for converter in converters:
                self._add_converter(converter)

        else:
            self._add_converter(converters)

    def __repr__(self) -> str:
        return f"{type(self).__name__} <{self._key}>"

    @property
    def converters(self) -> collections.Sequence[ConverterSig[typing.Any]]:
        """Sequence of the converters registered for this parameter."""
        return self._converters.copy()

    @property
    def default(self) -> typing.Any:
        """The parameter's default.

        If this is [tanjun.abc.NO_DEFAULT][] then this parameter is
        required.

        If this is [tanjun.abc.NO_PASS][] then the parameter won't be
        passed when no value was provided.
        """
        return self._default

    @property
    def is_multi(self) -> bool:
        """Whether this parameter is "multi".

        Multi parameters will be passed a list of all the values provided for
        this parameter (with each entry being converted separately.)
        """
        return self._is_multi

    @property
    def min_length(self) -> typing.Optional[int]:
        """If set, this parameters's parsed values will have to have lengths greater than or equal to this.

        If any converters are provided then this should be compatible with the
        result of them.
        """
        return self._min_length

    @property
    def max_length(self) -> typing.Optional[int]:
        """If set, this parameters's parsed values will have to have lengths less than or equal to this.

        If any converters are provided then this should be compatible with the
        result of them.
        """
        return self._max_length

    @property
    def min_value(self) -> typing.Optional[_CmpProto[typing.Any]]:
        """If set, this parameters's parsed values will have to be greater than or equal to this.

        If any converters are provided then this should be compatible with the
        result of them.
        """
        return self._min_value

    @property
    def max_value(self) -> typing.Optional[_CmpProto[typing.Any]]:
        """If set, this parameters's parsed values will have to be less than or equal to this.

        If any converters are provided then this should be compatible with the
        result of them.
        """
        return self._max_value

    @property
    def key(self) -> str:
        """The key of this parameter used to pass the result to the command's callback."""
        return self._key

    def _add_converter(self, converter: ConverterSig[typing.Any], /) -> None:
        if isinstance(converter, conversion.BaseConverter) and self._client:
            converter.check_client(self._client, f"{self._key} parameter")

        # Some types like `bool` and `bytes` are overridden here for the sake of convenience.
        converter = conversion.override_type(converter)
        self._converters.append(converter)

    def bind_client(self, client: tanjun.Client, /) -> None:
        self._client = client
        for converter in self._converters:
            if isinstance(converter, conversion.BaseConverter):
                converter.check_client(client, f"{self._key} parameter")

    def bind_component(self, component: tanjun.Component, /) -> None:
        self._component = component

    def _validate(self, value: typing.Any, /) -> None:
        # asserts value >= self._min_value
        if self._min_value is not None and self._min_value > value:
            raise errors.ConversionError(
                f"{self._key!r} must be greater than or equal to {self._min_value!r}", self._key
            )

        # asserts value <= self._max_value
        if self._max_value is not None and self._max_value < value:
            raise errors.ConversionError(f"{self._key!r} must be less than or equal to {self._max_value!r}", self._key)

        length: typing.Optional[int] = None
        # asserts that len(value) >= self._min_length
        if self._min_length is not None and self._min_length > (length := len(value)):
            raise errors.ConversionError(f"{self._key!r} must be longer than {self._min_length - 1}", self._key)

        # asserts that len(value) <= self._max_length
        if self._max_length is not None and self._max_length < (len(value) if length is None else length):
            raise errors.ConversionError(f"{self._key!r} can't be longer than {self._max_length}", self._key)

    async def convert(self, ctx: tanjun.Context, value: str, /) -> typing.Any:
        """Convert the given value to the type of this parameter."""
        if not self._converters:
            self._validate(value)
            return value

        sources: list[ValueError] = []
        for converter in self._converters:
            try:
                result = await ctx.call_with_async_di(converter, value)

            except ValueError as exc:
                sources.append(exc)

            else:
                self._validate(result)
                return result

        parameter_type = "option" if isinstance(self, Option) else "argument"
        raise errors.ConversionError(f"Couldn't convert {parameter_type} '{self._key}'", self._key, errors=sources)

    def copy(self) -> Self:
        """Copy the parameter.

        Returns
        -------
        Self
            A copy of the parameter.
        """
        inst = copy.copy(self)
        inst._converters = [copy.copy(converter) for converter in self._converters]
        return inst


class Argument(Parameter):
    """Representation of a positional argument used by the standard parser."""

    __slots__ = ("_is_greedy",)

    def __init__(
        self,
        key: str,
        /,
        *,
        converters: _MaybeIterable[ConverterSig[typing.Any]] = (),
        default: typing.Any = tanjun.NO_DEFAULT,
        greedy: bool = False,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_CmpProto[typing.Any]] = None,
        max_value: typing.Optional[_CmpProto[typing.Any]] = None,
        multi: bool = False,
    ) -> None:
        """Initialise a positional argument.

        Parameters
        ----------
        key
            The string identifier of this argument (may be used to pass the result
            of this argument to the command's callback during execution).
        converters
            The converter(s) this argument should use to handle values passed to it
            during parsing.

            If no converters are provided then the raw string value will be passed.

            Only the first converter to pass will be used.
        default
            The default value of this argument, if left as
            [tanjun.abc.NO_DEFAULT][] then this will have no default.

            If this is [tanjun.abc.NO_PASS][] then the `key` parameter won't be
            passed when no value was provided.
        greedy
            Whether or not this argument should be greedy (meaning that it
            takes in the remaining argument values).
        min_length
            Assert that a string argument's length is greater than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        max_length
            Assert that a string argument's length is less than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        min_value
            Assert that the parsed value(s) for this option are greater than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        max_value
            Assert that the parsed value(s) for this option are less than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.

        multi
            Whether this argument can be passed multiple times.
        """
        if greedy and multi:
            raise ValueError("Argument cannot be both greed and multi.")

        self._is_greedy = greedy
        super().__init__(
            key,
            converters=converters,
            default=default,
            min_length=min_length,
            max_length=max_length,
            min_value=min_value,
            max_value=max_value,
            multi=multi,
        )

    @property
    def is_greedy(self) -> bool:
        """Whether this parameter is greedy.

        Greedy parameters will consume the remaining message content as one
        string (with converters also being passed the whole string).

        !!! note
            Greedy and multi parameters cannot be used together.
        """
        return self._is_greedy


class Option(Parameter):
    """Representation of a named optional parameter used by the standard parser."""

    __slots__ = ("_empty_value", "_names")

    def __init__(
        self,
        key: str,
        name: str,
        /,
        *names: str,
        converters: _MaybeIterable[ConverterSig[typing.Any]] = (),
        default: typing.Any,
        empty_value: typing.Any = tanjun.NO_DEFAULT,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_CmpProto[typing.Any]] = None,
        max_value: typing.Optional[_CmpProto[typing.Any]] = None,
        multi: bool = True,
    ) -> None:
        """Initialise a named optional parameter.

        Parameters
        ----------
        key
            The string identifier of this option which will be used to pass the
            result of this argument to the command's callback during execution as
            a keyword argument.
        name
            The name of this option used for identifying it in the parsed content.
        *names
            Other names of this option used for identifying it in the parsed content.
        default
            The default value of this argument, unlike arguments this is required
            for options.
        converters
            The converter(s) this argument should use to handle values passed to it
            during parsing.

            If no converters are provided then the raw string value will be passed.

            Only the first converter to pass will be used.
        empty_value
            The value to use if this option is provided without a value.

            If left as [tanjun.abc.NO_DEFAULT][] then this option
            will error if it's provided without a value.

            [tanjun.abc.NO_PASS][] is not supported for this.
        min_length
            Assert that a string argument's length is greater than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        max_length
            Assert that a string argument's length is less than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        min_value
            Assert that the parsed value(s) for this option are greater than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.
        max_value
            Assert that the parsed value(s) for this option are less than or equal to this.

            If any converters are provided then this should be compatible
            with the result of them.

        multi
            If this option can be provided multiple times.
        """
        if not name.startswith("-") or not all(n.startswith("-") for n in names):
            raise ValueError("All option names must start with `-`")

        self._empty_value = empty_value
        self._names: list[str] = [name, *names]
        super().__init__(
            key,
            converters=converters,
            default=default,
            min_length=min_length,
            max_length=max_length,
            min_value=min_value,
            max_value=max_value,
            multi=multi,
        )

    @property
    def empty_value(self) -> typing.Any:
        """The value to return if the option is empty.

        If this is [tanjun.abc.NO_DEFAULT][] then a value will be
        required for the option.
        """
        return self._empty_value

    @property
    def names(self) -> collections.Sequence[str]:
        """Sequence of the CLI names of this option."""
        return self._names.copy()

    def __repr__(self) -> str:
        return f"{type(self).__name__} <{self.key}, {self._names}>"


class ShlexParser(AbstractOptionParser):
    """A shlex based [AbstractOptionParser][tanjun.parsing.AbstractOptionParser] implementation."""

    __slots__ = ("_arguments", "_callback_arg_names", "_client", "_component", "_options")

    def __init__(self) -> None:
        """Initialise a shlex parser."""
        self._arguments: list[Argument] = []
        self._callback_arg_names: list[tuple[str, collections.Container[str]]] = []
        self._client: typing.Optional[tanjun.Client] = None
        self._component: typing.Optional[tanjun.Component] = None
        self._options: list[Option] = []  # TODO: maybe switch to dict[str, Option] and assert doesn't already exist

    @property
    def arguments(self) -> collections.Sequence[Argument]:
        # <<inherited docstring from AbstractOptionParser>>.
        return self._arguments.copy()

    @property
    def options(self) -> collections.Sequence[Option]:
        # <<inherited docstring from AbstractOptionParser>>.
        return self._options.copy()

    def copy(self) -> Self:
        # <<inherited docstring from AbstractOptionParser>>.
        inst = copy.copy(self)
        inst._arguments = [argument.copy() for argument in self._arguments]
        inst._options = [option.copy() for option in self._options]
        return inst

    def _assert_key(self, key: str, /) -> None:
        for callback_name, names in self._callback_arg_names:
            if key not in names:
                raise ValueError(f"{key!r} is not a valid keyword argument for {callback_name}")

    @typing.overload
    def add_argument(
        self,
        key: str,
        /,
        converters: _MaybeIterable[ConverterSig[typing.Any]],
        *,
        default: typing.Any = tanjun.NO_DEFAULT,
        greedy: bool = False,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    def add_argument(
        self,
        key: str,
        /,
        converters: _MaybeIterable[ConverterSig[str]] = (),
        *,
        default: typing.Any = tanjun.NO_DEFAULT,
        greedy: bool = False,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_CmpProto[str]] = None,
        max_value: typing.Optional[_CmpProto[str]] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    def add_argument(
        self,
        key: str,
        /,
        converters: _MaybeIterable[ConverterSig[_SizedCmpProtoT]],
        *,
        default: typing.Any = tanjun.NO_DEFAULT,
        greedy: bool = False,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_SizedCmpProtoT] = None,
        max_value: typing.Optional[_SizedCmpProtoT] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    def add_argument(
        self,
        key: str,
        /,
        converters: _MaybeIterable[ConverterSig[collections.Sized]],
        *,
        default: typing.Any = tanjun.NO_DEFAULT,
        greedy: bool = False,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    def add_argument(
        self,
        key: str,
        /,
        converters: _MaybeIterable[ConverterSig[_CmpProtoT]],
        *,
        default: typing.Any = tanjun.NO_DEFAULT,
        greedy: bool = False,
        min_value: typing.Optional[_CmpProtoT] = None,
        max_value: typing.Optional[_CmpProtoT] = None,
        multi: bool = False,
    ) -> Self:
        ...

    def add_argument(
        self,
        key: str,
        /,
        converters: _MaybeIterable[ConverterSig[typing.Any]] = (),
        *,
        default: typing.Any = tanjun.NO_DEFAULT,
        greedy: bool = False,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_CmpProto[typing.Any]] = None,
        max_value: typing.Optional[_CmpProto[typing.Any]] = None,
        multi: bool = False,
    ) -> Self:
        # <<inherited docstring from AbstractOptionParser>>.
        self._assert_key(key)
        argument = Argument(
            key,
            converters=converters,
            default=default,
            greedy=greedy,
            min_length=min_length,
            max_length=max_length,
            min_value=min_value,
            max_value=max_value,
            multi=multi,
        )
        if self._client:
            argument.bind_client(self._client)

        if self._component:
            argument.bind_component(self._component)

        for argument_ in self._arguments:
            if argument_.is_multi or argument_.is_greedy:
                raise ValueError("Multi or greedy argument must be the last argument")

        self._arguments.append(argument)
        return self

    @typing.overload
    def add_option(
        self,
        key: str,
        name: str,
        /,
        *names: str,
        converters: _MaybeIterable[ConverterSig[typing.Any]],
        default: typing.Any,
        empty_value: typing.Any = tanjun.NO_DEFAULT,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    def add_option(
        self,
        key: str,
        name: str,
        /,
        *names: str,
        converters: _MaybeIterable[ConverterSig[str]] = (),
        default: typing.Any,
        empty_value: typing.Any = tanjun.NO_DEFAULT,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_CmpProto[str]] = None,
        max_value: typing.Optional[_CmpProto[str]] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    def add_option(
        self,
        key: str,
        name: str,
        /,
        *names: str,
        converters: _MaybeIterable[ConverterSig[_SizedCmpProtoT]],
        default: typing.Any,
        empty_value: typing.Any = tanjun.NO_DEFAULT,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_SizedCmpProtoT] = None,
        max_value: typing.Optional[_SizedCmpProtoT] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    def add_option(
        self,
        key: str,
        name: str,
        /,
        *names: str,
        converters: _MaybeIterable[ConverterSig[collections.Sized]],
        default: typing.Any,
        empty_value: typing.Any = tanjun.NO_DEFAULT,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        multi: bool = False,
    ) -> Self:
        ...

    @typing.overload
    def add_option(
        self,
        key: str,
        name: str,
        /,
        *names: str,
        converters: _MaybeIterable[ConverterSig[_CmpProtoT]],
        default: typing.Any,
        empty_value: typing.Any = tanjun.NO_DEFAULT,
        min_value: typing.Optional[_CmpProtoT] = None,
        max_value: typing.Optional[_CmpProtoT] = None,
        multi: bool = False,
    ) -> Self:
        ...

    # TODO: add default getter
    def add_option(
        self,
        key: str,
        name: str,
        /,
        *names: str,
        converters: _MaybeIterable[ConverterSig[typing.Any]] = (),
        default: typing.Any,
        empty_value: typing.Any = tanjun.NO_DEFAULT,
        min_length: typing.Optional[int] = None,
        max_length: typing.Optional[int] = None,
        min_value: typing.Optional[_CmpProto[typing.Any]] = None,
        max_value: typing.Optional[_CmpProto[typing.Any]] = None,
        multi: bool = False,
    ) -> Self:
        # <<inherited docstring from AbstractOptionParser>>.
        self._assert_key(key)
        option = Option(
            key,
            name,
            *names,
            converters=converters,
            default=default,
            empty_value=empty_value,
            min_length=min_length,
            max_length=max_length,
            min_value=min_value,
            max_value=max_value,
            multi=multi,
        )

        if self._client:
            option.bind_client(self._client)

        if self._component:
            option.bind_component(self._component)

        self._options.append(option)
        return self

    def bind_client(self, client: tanjun.Client, /) -> Self:
        # <<inherited docstring from AbstractOptionParser>>.
        self._client = client
        for parameter in itertools.chain(self._options, self._arguments):
            parameter.bind_client(client)

        return self

    def bind_component(self, component: tanjun.Component, /) -> Self:
        # <<inherited docstring from AbstractOptionParser>>.
        self._component = component
        for parameter in itertools.chain(self._options, self._arguments):
            parameter.bind_component(component)

        return self

    def parse(
        self, ctx: tanjun.MessageContext, /
    ) -> collections.Coroutine[typing.Any, typing.Any, dict[str, typing.Any]]:
        # <<inherited docstring from AbstractOptionParser>>.
        return _SemanticShlex(ctx, self._arguments, self._options).parse()

    def validate_arg_keys(self, callback_name: str, names: collections.Container[str], /) -> None:
        # <<inherited docstring from AbstractOptionParser>>.
        self._callback_arg_names.append((callback_name, names))

        for parameter in itertools.chain(self._options, self._arguments):
            if parameter.key not in names:
                raise ValueError(f"{parameter.key!r} is not a valid keyword argument for {callback_name}")


def with_parser(command: _CommandT, /) -> _CommandT:
    """Add a shlex parser command parser to a supported command.

    Example
    -------
    ```py
    @tanjun.with_argument("arg", converters=int)
    @tanjun.with_parser
    @tanjun.as_message_command("hi")
    async def hi(ctx: tanjun.MessageContext, arg: int) -> None:
        ...
    ```

    Parameters
    ----------
    command
        The message command to set the parser on.

    Returns
    -------
    tanjun.abc.MessageCommand
        The command with the parser set.

    Raises
    ------
    ValueError
        If the command already has a parser set.
    """
    if command.parser:
        raise ValueError("Command already has a parser set")

    return command.set_parser(ShlexParser())



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Utility classes for making cache calls."""
from __future__ import annotations

import typing

import hikari

from ..dependencies import async_cache

if typing.TYPE_CHECKING:
    from .. import abc as tanjun

_ChannelCacheT = async_cache.SfCache[hikari.PermissibleGuildChannel]
_ThreadCacheT = async_cache.SfCache[hikari.GuildThreadChannel]
_THREAD_CHANNEL_TYPES = frozenset(
    (
        hikari.ChannelType.GUILD_NEWS_THREAD,
        hikari.ChannelType.GUILD_PUBLIC_THREAD,
        hikari.ChannelType.GUILD_PRIVATE_THREAD,
    )
)


async def get_perm_channel(client: tanjun.Client, channel_id: hikari.Snowflake, /) -> hikari.PermissibleGuildChannel:
    """Get the permissionable channel for a channel.

    This will resolve threads to their parent channel.

    Parameters
    ----------
    client
        The client to use to get the channel.
    channel_id
        ID of the target channel

    Returns
    -------
    hikari.channels.PermissibleGuildChannel
        The permissible guild channel.
    """
    channel: typing.Optional[hikari.PartialChannel]  # MyPy compat
    if client.cache and (channel := client.cache.get_guild_channel(channel_id)):
        return channel

    thread_cache = client.injector.get_type_dependency(_ThreadCacheT)
    if thread_cache and (thread := await thread_cache.get(channel_id, default=None)):
        if client.cache and (channel := client.cache.get_guild_channel(thread.parent_id)):
            return channel

        channel_id = thread.parent_id

    channel_cache = client.injector.get_type_dependency(_ChannelCacheT, default=None)
    if channel_cache and (channel := await channel_cache.get(channel_id, default=None)):
        return channel

    channel = await client.rest.fetch_channel(channel_id)
    if channel.type not in _THREAD_CHANNEL_TYPES:
        assert isinstance(channel, hikari.PermissibleGuildChannel)
        return channel

    assert isinstance(channel, hikari.GuildChannel)
    assert channel.parent_id is not None
    if client.cache and (channel_ := client.cache.get_guild_channel(channel.parent_id)):
        return channel_

    if channel_cache and (channel_ := await channel_cache.get(channel.parent_id, default=None)):
        return channel_

    channel = await client.rest.fetch_channel(channel.parent_id)
    assert isinstance(channel, hikari.PermissibleGuildChannel)
    return channel



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportPrivateUsage=none

from unittest import mock

import alluka

import tanjun


def test_as_self_injecting():
    mock_callback = mock.Mock()
    mock_client = mock.Mock()

    result = tanjun.injecting.as_self_injecting(mock_client)(mock_callback)

    assert result.callback is mock_callback
    assert result._client is mock_client.injector


def test_aliases():
    assert set(tanjun.injecting.__all__) == {
        "AbstractInjectionContext",
        "BasicInjectionContext",
        "CallbackSig",
        "Injected",
        "InjectorClient",
        "SelfInjectingCallback",
        "UNDEFINED",
        "Undefined",
        "UndefinedOr",
        "as_self_injecting",
        "inject",
        "injected",
    }
    assert tanjun.injecting.SelfInjectingCallback is alluka.AsyncSelfInjecting
    assert tanjun.injecting.BasicInjectionContext is alluka.BasicContext
    assert tanjun.injecting.InjectorClient is alluka.Client
    assert tanjun.injecting.Injected is alluka.Injected
    assert tanjun.injecting.inject is alluka.inject
    assert tanjun.injecting.injected is alluka.inject
    assert tanjun.injecting.UNDEFINED is alluka.abc.UNDEFINED
    assert tanjun.injecting.CallbackSig is alluka.abc.CallbackSig
    assert tanjun.injecting.AbstractInjectionContext is alluka.abc.Context
    assert tanjun.injecting.Undefined is alluka.abc.Undefined



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import contextlib
import dataclasses
import datetime
import enum
import hmac
import io
import logging
import os
import pathlib
import subprocess  # noqa: S404
import sys
import tempfile
import time
import traceback
import types
import typing
import zipfile
from collections import abc as collections
from typing import Self

import anyio
import dotenv
import fastapi
import httpx
import jwt
import piped_shared
import starlette.middleware
from anyio.streams import memory as streams
from asgiref import typing as asgiref
from dateutil import parser as dateutil

if sys.version_info < (3, 11):
    raise RuntimeError("Only supports Python 3.11+")

_LOGGER = logging.getLogger("piped.bot")
_LOGGER.setLevel("INFO")

dotenv.load_dotenv()

_username = os.environ.get("CLIENT_NAME", default="always-on-duty") + "[bot]"

with httpx.Client() as client:
    _user_id = int(client.get(f"https://api.github.com/users/{_username}").json()["id"])

APP_ID = os.environ["APP_ID"]
COMMIT_ENV = {
    "GIT_AUTHOR_NAME": _username,
    "GIT_AUTHOR_EMAIL": f"{_user_id}+{_username}@users.noreply.github.com",
    "GIT_COMMITTER_NAME": _username,
}
COMMIT_ENV["GIT_COMMITTER_EMAIL"] = COMMIT_ENV["GIT_AUTHOR_EMAIL"]
CLIENT_SECRET = os.environ["CLIENT_SECRET"].encode()
WEBHOOK_SECRET = os.environ["WEBHOOK_SECRET"]
jwt_instance = jwt.JWT()


class _ProcessingIndex:
    """Index of the PRs being processed."""

    __slots__ = ("_prs", "_repos")

    def __init__(self) -> None:
        self._prs: dict[str, anyio.CancelScope] = {}
        self._repos: dict[int, list[int]] = {}

    def _pr_id(self, repo_id: int, pr_id: int, /) -> str:
        return f"{repo_id}:{pr_id}"

    async def close(self) -> None:
        """Cancel all active PR processing tasks."""
        self._repos.clear()
        prs = self._prs
        self._prs = {}
        _LOGGER.info("Stopping all current calls")
        for scope in prs.values():
            scope.cancel()

        await anyio.sleep(0)  # Yield to the loop to let these cancels propagate

    def clear_for_repo(self, repo_id: int, /, *, repo_name: str | None = None) -> None:
        """Cancel the active PR processing tasks for a Repo.

        Parameters
        ----------
        repo_id
            ID of the repo to cancel the active tasks for.
        repo_name
            Name of the repo to cancel the active tasks for.

            Used for logging.
        """
        if prs := self._repos.get(repo_id):
            repo_name = repo_name or str(repo_id)
            _LOGGER.info("Stopping calls for all PRs in %s", repo_name)
            for pr_id in prs:
                self._prs.pop(self._pr_id(repo_id, pr_id)).cancel()

    def stop_for_pr(self, repo_id: int, pr_id: int, /, *, repo_name: str | None = None) -> None:
        """Cancel the active processing task for a PR.

        Parameters
        ----------
        repo_id
            ID of the repo to cancel an active task in.
        pr_id
            ID of the PR to cancel the active processing task for.
        repo_name
            Name of the repo to cancel an active task in.

            Use for logging.
        """
        if pr := self._prs.pop(self._pr_id(repo_id, pr_id), None):
            repo_name = repo_name or str(repo_id)
            _LOGGER.info("Stopping call for %s:%s", repo_name, pr_id)
            pr.cancel()

    def start(self, repo_id: int, pr_id: int, /, *, repo_name: str | None = None) -> anyio.CancelScope:
        """Create a cancel scope for processing a specific PR.

        This will cancel any previous processing task for the passed PR.

        Parameters
        ----------
        repo_id
            ID of the repo to start a processing task for.
        pr_id
            ID of the pull request to start a processing task for.
        repo_name
            Name of the repo this task is in.

            Use for logging.

        Returns
        -------
        anyio.CancelScope
            Cancel scope to use for this task.
        """
        key = self._pr_id(repo_id, pr_id)
        repo_name = repo_name or str(repo_id)

        if pr := self._prs.pop(key, None):
            _LOGGER.info("Stopping previous call for for %s:%s", repo_name, pr_id)
            pr.cancel()

        _LOGGER.info("Starting call for %s:%s", repo_name, pr_id)
        scope = anyio.CancelScope()
        self._prs[key] = scope
        return scope


class _Tokens:
    """Index of the Github API tokens this application has authorised."""

    __slots__ = ("_installation_tokens", "_private_key")

    def __init__(self) -> None:
        self._installation_tokens: dict[int, tuple[datetime.datetime, str]] = {}
        self._private_key = jwt.jwk_from_pem(os.environ["PRIVATE_KEY"].encode())

    def app_token(self, *, on_gen: collections.Callable[[str], None] | None = None) -> str:
        """Generate an application app token.

        !!! warning
            This cannot does not provide authorization for repos or
            organisations the application is authorised in.

        Parameters
        ----------
        on_gen
            Called on new token generation.

            This is for log filtering.
        """
        now = int(time.time())
        token = jwt_instance.encode(
            {"iat": now - 60, "exp": now + 60 * 2, "iss": APP_ID}, self._private_key, alg="RS256"
        )

        if on_gen:
            on_gen(token)

        return token

    async def installation_token(
        self,
        http: httpx.AsyncClient,
        installation_id: int,
        /,
        *,
        on_gen: collections.Callable[[str], None] | None = None,
    ) -> str:
        """Authorise an installation specific token.

        This is used to authorise organisation and repo actions and will return
        cached tokens.

        Parameters
        ----------
        http
            REST client to use to authorise the token.
        installation_id
            ID of the installation to authorise a token for.
        on_gen
            Called on new token generation for both app and installation tokens.

            This is for log filtering.

        Returns
        -------
        str
            The generated installation token.
        """
        if token_info := self._installation_tokens.get(installation_id):
            expire_by = datetime.datetime.now(tz=datetime.timezone.utc) - datetime.timedelta(seconds=60)
            if token_info[0] >= (expire_by):
                return token_info[1]

        # TODO: do we need/want to use an async lock here?
        response = await _request(
            http,
            "POST",
            f"/app/installations/{installation_id}/access_tokens",
            json={"permissions": {"actions": "read", "checks": "write", "contents": "write", "workflows": "write"}},
            token=self.app_token(on_gen=on_gen),
        )
        data = response.json()
        token: str = data["token"]
        if on_gen:
            on_gen(token)

        self._installation_tokens[installation_id] = (dateutil.isoparse(data["expires_at"]), token)
        return token


async def _request(
    http: httpx.AsyncClient,
    method: typing.Literal["GET", "PATCH", "POST", "DELETE"],
    endpoint: str,
    /,
    *,
    json: dict[str, typing.Any] | None = None,
    query: dict[str, str] | None = None,
    output: typing.IO[str] | None = None,
    token: str | None = None,
) -> httpx.Response:
    """Make a request to Github's API.

    Parameters
    ---------
    http
        The REST client to use to make the request.
    endpoint
        Endpoint to request to.

        This will be appended to `"https://api.github.com"` if it doesn't
        start with `"https://"`.
    json
        Dict of the JSON payload to include in this request.
    query
        Dict of the query string parameters to include for this request.
    token
        The authorisation token to use.
    """
    if not endpoint.startswith("https://"):
        endpoint = f"https://api.github.com{endpoint}"

    headers = {"X-GitHub-Api-Version": "2022-11-28"}
    if token:
        headers["Authorization"] = f"Bearer {token}"

    if json:
        headers["Content-Type"] = "application/vnd.github+json"

    response = await http.request(method, endpoint, follow_redirects=True, headers=headers, json=json, params=query)

    try:
        response.raise_for_status()
    except Exception:
        print("Response body:", file=output or sys.stderr)  # noqa: T201
        print(response.read().decode(), file=output or sys.stderr)  # noqa: T201
        raise

    return response


class _CachedReceive:
    """Helper ASGI event receiver which first returned the cached request body."""

    __slots__ = ("_data", "_receive")

    def __init__(self, data: bytearray, receive: asgiref.ASGIReceiveCallable) -> None:
        self._data: bytearray | None = data  # TODO: should this be chunked?
        self._receive = receive  # TODO: check this behaviour

    async def __call__(self) -> asgiref.ASGIReceiveEvent:
        if not self._data:
            return await self._receive()

        data = self._data
        self._data = None
        return {"type": "http.request", "body": data, "more_body": False}


async def _error_response(send: asgiref.ASGISendCallable, body: bytes, /, *, status_code: int = 400) -> None:
    """Helper function for returning a quick error response.

    Parameters
    ----------
    send
        The ASGI send callback to use to send this response.
    body
        The error message.
    status_code
        The error's status code.
    """
    await send(
        {
            "type": "http.response.start",
            "status": status_code,
            "headers": [(b"content-type", b"text/plain; charset=UTF-8")],
            "trailers": False,
        }
    )
    await send({"type": "http.response.body", "body": body, "more_body": False})


def _find_headers(scope: asgiref.HTTPScope, headers: collections.Collection[bytes]) -> dict[bytes, bytes]:
    """Helper function for extracting specific headers from an ASGI request.

    Parameters
    ----------
    scope
        The ASGI HTTP scope payload to get the headers from.
    headers
        Collection of the headers to find.
    """
    results: dict[bytes, bytes] = {}

    for header_name, header_value in scope["headers"]:
        header_name = header_name.lower()
        if header_name in headers:
            results[header_name] = header_value

            if len(results) == len(headers):
                break

    return results


# TODO: check user agent header starts with "GitHub-Hookshot/"?
class AuthMiddleware:
    """ASGI signature authorisation middleware."""

    __slots__ = ("app",)

    # starlette.types.ASGIApp is more appropriate but less concise than this callable type.
    def __init__(
        self,
        app: collections.Callable[
            [asgiref.Scope, asgiref.ASGIReceiveCallable, asgiref.ASGISendCallable], collections.Awaitable[None]
        ],
    ) -> None:
        self.app = app

    async def __call__(
        self, scope: asgiref.Scope, receive: asgiref.ASGIReceiveCallable, send: asgiref.ASGISendCallable
    ) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        signature = _find_headers(scope, (b"x-hub-signature-256",)).get(b"x-hub-signature-256")
        if not signature:
            await _error_response(send, b"Missing signature header")
            return

        more_body = True
        payload = bytearray()
        while more_body:
            event = await receive()
            match event:
                case {"type": "http.request"}:
                    more_body = event.get("more_body", False)
                    payload.extend(event.get("body") or b"")
                case _:
                    raise NotImplementedError

        digest = "sha256=" + hmac.new(CLIENT_SECRET, payload, digestmod="sha256").hexdigest()

        if not hmac.compare_digest(signature.decode(), digest):
            await _error_response(send, b"Invalid signature", status_code=401)
            return

        await self.app(scope, _CachedReceive(payload, receive), send)


@dataclasses.dataclass(slots=True)
class _Workflow:
    name: str
    workflow_id: int


class _WorkflowAction(str, enum.Enum):
    COMPLETED = "completed"
    IN_PROGRESS = "in_progress"
    REQUESTED = "requested"


class _WorkflowDispatch:
    """Workflow dispatch tracker."""

    __slots__ = ("_listeners",)

    def __init__(self) -> None:
        self._listeners: dict[
            tuple[int, int, str], streams.MemoryObjectSendStream[tuple[int, str, _WorkflowAction]]
        ] = {}

    def consume_event(self, body: dict[str, typing.Any], /) -> None:
        """Dispatch a workflow_run event.

        Parameters
        ----------
        body
            Dict body of the workflow_run event.
        """
        head_repo_id = int(body["workflow_run"]["head_repository"]["id"])
        head_sha: str = body["workflow_run"]["head_sha"]
        repo_id = int(body["repository"]["id"])

        if send := self._listeners.get((repo_id, head_repo_id, head_sha)):
            action = _WorkflowAction(body["action"])
            name: str = body["workflow_run"]["name"]
            workflow_id = int(body["workflow_run"]["id"])
            send.send_nowait((workflow_id, name, action))

    @contextlib.contextmanager
    def track_workflows(
        self, repo_id: int, head_repo_id: int, head_sha: str, /
    ) -> collections.Generator["_IterWorkflows", None, None]:
        """Async context manager which manages tracking the workflows for a PR.

        Parameters
        ----------

        Returns
        -------
        _IterWorkflows
            Async iterable of the received workflow finishes.

            `_IterWorkflows.filter_names` should be used to set this to filter
            for specific names before iterating over this.
        """
        key = (repo_id, head_repo_id, head_sha)
        send, recv = anyio.create_memory_object_stream(1_000, item_type=tuple[int, str, _WorkflowAction])
        self._listeners[key] = send

        yield _IterWorkflows(recv)

        send.close()
        del self._listeners[key]


class _IterWorkflows:
    """Async iterable of received workflow finishes."""

    __slots__ = ("_filter", "_recv")

    def __init__(self, recv: streams.MemoryObjectReceiveStream[tuple[int, str, _WorkflowAction]], /) -> None:
        self._filter: collections.Collection[str] = ()
        self._recv = recv

    async def __aiter__(self) -> collections.AsyncIterator[_Workflow]:
        timeout_at = time.time() + 5
        waiting_on = {name: False for name in self._filter}

        while waiting_on:
            if not any(waiting_on.values()):
                if time.time() > timeout_at:
                    break

                timeout = timeout_at

            else:
                timeout = None

            try:
                with anyio.fail_after(timeout):
                    workflow_id, name, action = await self._recv.receive()

            except TimeoutError:
                return

            if name not in waiting_on:
                continue

            if action is _WorkflowAction.COMPLETED:
                del waiting_on[name]
                yield _Workflow(name, workflow_id)

            else:
                waiting_on[name] = True

    def filter_names(self, names: collections.Collection[str], /) -> Self:
        """Set this to only track specific workflows.

        This will override any previously set filter.

        Parameters
        ----------
        names
            Collection of workflow names to filter for.

        Returns
        -------
        typing.Self
            The async workflow iterable.
        """
        self._filter = names
        return self


async def _on_startup():
    app.state.http = httpx.AsyncClient()
    app.state.index = _ProcessingIndex()
    app.state.tokens = _Tokens()
    app.state.workflows = _WorkflowDispatch()


async def _on_shutdown():
    assert isinstance(app.state.index, _ProcessingIndex)
    assert isinstance(app.state.http, httpx.AsyncClient)
    await app.state.index.close()
    await app.state.http.aclose()


app = fastapi.FastAPI(middleware=[starlette.middleware.Middleware(AuthMiddleware)])
app.router.on_startup.append(_on_startup)
app.router.on_shutdown.append(_on_shutdown)


@app.post("/webhook")
async def post_webhook(
    body: dict[str, typing.Any],
    request: fastapi.Request,
    tasks: fastapi.BackgroundTasks,
    x_github_event: str = fastapi.Header(),
) -> fastapi.Response:
    assert isinstance(request.app.state.http, httpx.AsyncClient)
    assert isinstance(request.app.state.index, _ProcessingIndex)
    assert isinstance(request.app.state.tokens, _Tokens)
    assert isinstance(request.app.state.workflows, _WorkflowDispatch)
    http = request.app.state.http
    index = request.app.state.index
    tokens = request.app.state.tokens
    workflows = request.app.state.workflows
    match (x_github_event, body):
        case ("pull_request", {"action": "closed", "number": number, "repository": repo_data}):
            index.stop_for_pr(int(repo_data["id"]), int(number), repo_name=repo_data["full_name"])
            await anyio.sleep(0)  # Yield to the loop to let these cancels propagate

        case ("pull_request", {"action": "opened" | "reopened" | "synchronize"}):
            tasks.add_task(_process_repo, http, tokens, index, workflows, body)
            return fastapi.Response(status_code=202)

        case ("workflow_run", _):
            workflows.consume_event(body)

        case ("installation", {"action": "removed", "repositories_removed": repositories_removed}):
            for repo in repositories_removed:
                index.clear_for_repo(int(repo["id"]))

            await anyio.sleep(0)  # Yield to the loop to let these cancels propagate

        case ("installation_repositories", {"action": "removed", "repositories": repositories}):
            for repo in repositories:
                index.clear_for_repo(int(repo["id"]))

            await anyio.sleep(0)  # Yield to the loop to let these cancels propagate

        # Guard to let these expected but ignored cases still return 204
        case (
            # check_suite events are received for the bot's check suites
            # regardless of configuration.
            "check_suite"
            | "github_app_authorization"
            | "installation"
            | "installation_repositories"
            | "installation_target"
            | "pull_request"
            | "workflow_run",
            _,
        ):
            pass

        case _:
            _LOGGER.info(
                "Ignoring unexpected event type %r. These events should be disabled for this app", x_github_event
            )
            return fastapi.Response("Event type not implemented", status_code=501)

    return fastapi.Response(status_code=204)


@contextlib.asynccontextmanager
async def _with_cloned(
    output: typing.IO[str], url: str, /, *, branch: str = "master"
) -> collections.AsyncGenerator[pathlib.Path, None]:
    """Async context manager which shallow clones a repo into a temporary directory.

    Parameters
    ----------
    output
        String file-like object this should pipe GIT's output to.
    url
        URL of the repository to clone.

        This must include an installation which is authorised for
        `contents: write`.
        (`https://x-access-token:<token>@github.com/owner/repo.git`)
    branch
        The branch to clone.
    """
    temp_dir = await anyio.to_thread.run_sync(lambda: tempfile.TemporaryDirectory[str](ignore_cleanup_errors=True))
    try:
        await run_process(output, ["git", "clone", url, "--depth", "1", "--branch", branch, temp_dir.name])
        yield pathlib.Path(temp_dir.name)

    finally:
        # TODO: this just fails on Windows sometimes
        with anyio.CancelScope(shield=True):
            await anyio.to_thread.run_sync(temp_dir.cleanup)


class _RunCheck:
    """Context manager which manages the Github check suite for this application."""

    __slots__ = ("_check_id", "_commit_hash", "_filter_from_logs", "_http", "_output", "_repo_name", "_token")

    def __init__(self, http: httpx.AsyncClient, /, *, token: str, repo_name: str, commit_hash: str) -> None:
        """Initialise this context manager.

        Parameters
        ----------
        http
            REST client to use to manage the check suite.
        token
            Installation token to use to authorise the check suite requests.

            This must be authorised for `checks: write`.
        repo_name
            The repo's full name in the format `"{owner_name}/{repo_name}"`.
        commit_hash
            Hash of the PR commit this run is for.
        """
        self._check_id = -1
        self._commit_hash = commit_hash
        self._filter_from_logs = [token]
        self._http = http
        self._output = io.StringIO()
        self._repo_name = repo_name
        self._token = token

    @property
    def output(self) -> io.StringIO:
        return self._output

    async def __aenter__(self) -> Self:
        result = await _request(
            self._http,
            "POST",
            f"/repos/{self._repo_name}/check-runs",
            json={"name": "Inspecting PR", "head_sha": self._commit_hash},
            output=self._output,
            token=self._token,
        )
        self._check_id = int(result.json()["id"])
        return self

    async def __aexit__(
        self,
        exc_cls: type[BaseException] | None,
        exc: BaseException | None,
        traceback_value: types.TracebackType | None,
    ) -> None:
        self._output.seek(0)
        output = {}

        if exc:
            conclusion = "failure" if isinstance(exc, Exception) else "cancelled"
            output["title"] = "Error"
            output["summary"] = str(exc)
            self._output.write("\n")
            self._output.write("```python\n")
            traceback.print_exception(exc_cls, exc, traceback_value, file=self._output)
            self._output.write("```\n")

        else:
            conclusion = "success"
            output["title"] = output["summary"] = "Success"

        # TODO: charlimit
        text = "\n".join(_censor(line, self._filter_from_logs) for line in self._output)
        output["text"] = f"```\n{text}\n```"

        # TODO: https://docs.github.com/en/get-started/writing-on-github/
        # working-with-advanced-formatting/creating-and-highlighting-code-blocks
        with anyio.CancelScope(shield=True):
            await _request(
                self._http,
                "PATCH",
                f"/repos/{self._repo_name}/check-runs/{self._check_id}",
                json={"conclusion": conclusion, "output": output},
                output=self._output,
                token=self._token,
            )

    def filter_from_logs(self, value: str, /) -> Self:
        """Mark a string as being filtered out of the logs.

        Parameters
        ----------
        value
            String to censor from logs.
        """
        self._filter_from_logs.append(value)
        return self

    async def mark_running(self) -> None:
        """Mark the check suite as running.

        Raises
        ------
        RuntimeError
            If called outside of this context manager's context.
        """
        if self._check_id == -1:
            raise RuntimeError("Not running yet")

        await _request(
            self._http,
            "PATCH",
            f"/repos/{self._repo_name}/check-runs/{self._check_id}",
            json={"started_at": datetime.datetime.now(tz=datetime.timezone.utc).isoformat(), "status": "in_progress"},
            output=self._output,
            token=self._token,
        )


def _censor(value: str, filters: list[str], /) -> str:
    for filter_ in filters:
        value.replace(filter_, "***")

    return value


async def _process_repo(
    http: httpx.AsyncClient,
    tokens: _Tokens,
    index: _ProcessingIndex,
    workflows: _WorkflowDispatch,
    body: dict[str, typing.Any],
) -> None:
    repo_data: dict[str, typing.Any] = body["repository"]
    repo_id = int(repo_data["id"])
    pr_id = int(body["number"])
    full_name: str = repo_data["full_name"]
    head_name: str = body["pull_request"]["head"]["repo"]["full_name"]
    head_ref: str = body["pull_request"]["head"]["ref"]
    head_repo_id = int(body["pull_request"]["head"]["repo"]["id"])
    head_sha: str = body["pull_request"]["head"]["sha"]
    installation_id = int(body["installation"]["id"])

    with (
        index.start(repo_id, pr_id, repo_name=full_name),
        workflows.track_workflows(repo_id, head_repo_id, head_sha) as tracked_workflows,
    ):
        token = await tokens.installation_token(http, installation_id)
        git_url = f"https://x-access-token:{token}@github.com/{head_name}.git"
        _LOGGER.info("Cloning %s:%s branch %s", full_name, pr_id, head_ref)
        run_ctx = _RunCheck(http, token=token, repo_name=full_name, commit_hash=head_sha)

        async with run_ctx, _with_cloned(run_ctx.output, git_url, branch=head_ref) as temp_dir_path:
            config = await piped_shared.Config.read_async(temp_dir_path)
            if not config.bot_actions:
                _LOGGER.warn("Received event from %s repo with no bot_wait_for", full_name)
                return

            await run_ctx.mark_running()
            async for workflow in tracked_workflows.filter_names(config.bot_actions):
                await _apply_patch(http, run_ctx.output, token, full_name, workflow, cwd=temp_dir_path)

            await run_process(run_ctx.output, ["git", "push"], cwd=temp_dir_path)


async def _apply_patch(
    http: httpx.AsyncClient,
    output: typing.IO[str],
    token: str,
    repo_name: str,
    workflow: _Workflow,
    /,
    *,
    cwd: pathlib.Path,
) -> None:
    """Apply a patch file from another workflow's artifacts and commit its changes.

    This specifically looks for an artefact called `gogo.patch` and unzips it
    to get the file at `./gogo.patch`.

    Parameters
    ----------
    http
        The REST client to use to scan and download the workflow's artefacts.
    output
        String file-like object this should pipe GIT's output to.
    token
        The integration token to use.

        This must be authorised for `actions: read`.
    repo_name
        The repo's full name in the format `"{owner_name}/{repo_name}"`.
    workflow
        The workflow run to apply the patch of (if set).
    cwd
        Path to the target repo's top level directory.
    """
    # TODO: pagination support
    response = await _request(
        http,
        "GET",
        f"https://api.github.com/repos/{repo_name}/actions/runs/{workflow.workflow_id}/artifacts",
        output=output,
        query={"per_page": "100"},
        token=token,
    )
    for artifact in response.json()["artifacts"]:
        if artifact["name"] != "gogo.patch":
            continue

        response = await _request(http, "GET", artifact["archive_download_url"], output=output, token=token)
        zipped = zipfile.ZipFile(io.BytesIO(await response.aread()))
        # It's safe to extract to cwd since gogo.patch is git ignored.
        patch_path = await anyio.to_thread.run_sync(zipped.extract, "gogo.patch", cwd)

        try:
            # TODO: could --3way or --unidiff-zero help with conflicts here?
            await run_process(output, ["git", "apply", patch_path], cwd=cwd)

        # If this conflicted then we should allow another CI run to redo these
        # changes after the current changes have been pushed.
        except subprocess.CalledProcessError:
            await anyio.to_thread.run_sync(pathlib.Path(patch_path).unlink)

        else:
            await anyio.to_thread.run_sync(pathlib.Path(patch_path).unlink)
            await run_process(output, ["git", "add", "."], cwd=cwd, env=COMMIT_ENV)
            await run_process(output, ["git", "commit", "-am", workflow.name], cwd=cwd, env=COMMIT_ENV)

        break


async def run_process(
    output: typing.IO[str],
    command: str | bytes | collections.Sequence[str | bytes],
    *,
    input: bytes | None = None,  # noqa: A002
    check: bool = True,
    cwd: str | bytes | os.PathLike[str] | None = None,
    env: collections.Mapping[str, str] | None = None,
    start_new_session: bool = False,
) -> None:
    try:
        # TODO: could --3way or --unidiff-zero help with conflicts here?
        result = await anyio.run_process(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            input=input,
            check=check,
            cwd=cwd,
            env=env,
            start_new_session=start_new_session,
        )

    except subprocess.CalledProcessError as exc:
        assert isinstance(exc.stdout, bytes)
        output.write(exc.stdout.decode())
        raise

    else:
        output.write(result.stdout.decode())



# -*- coding: utf-8 -*-
# Tanjun Examples - A collection of examples for Tanjun.
# Written in 2021 by Lucina Lucina@lmbyrne.dev
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain worldwide.
# This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along with this software.
# If not, see <https://creativecommons.org/publicdomain/zero/1.0/>.
"""General examples how Tanjun's used."""
